"""Database models used by the SQLAlchemy result store backend."""
from datetime import datetime

import sqlalchemy as sa
from sqlalchemy.types import PickleType

from celery import states

from .session import ResultModelBase

__all__ = ('Task', 'TaskExtended', 'TaskSet')


class Task(ResultModelBase):
    """Task result/status."""

    __tablename__ = 'celery_taskmeta'
    __table_args__ = {'sqlite_autoincrement': True}

    id = sa.Column(sa.Integer, sa.Sequence('task_id_sequence'),
                   primary_key=True, autoincrement=True)
    task_id = sa.Column(sa.String(155), unique=True)
    status = sa.Column(sa.String(50), default=states.PENDING)
    result = sa.Column(PickleType, nullable=True)
    date_done = sa.Column(sa.DateTime, default=datetime.utcnow,
                          onupdate=datetime.utcnow, nullable=True)
    traceback = sa.Column(sa.Text, nullable=True)

    def __init__(self, task_id):
        self.task_id = task_id

    def to_dict(self):
        return {
            'task_id': self.task_id,
            'status': self.status,
            'result': self.result,
            'traceback': self.traceback,
            'date_done': self.date_done,
        }

    def __repr__(self):
        return '<Task {0.task_id} state: {0.status}>'.format(self)

    @classmethod
    def configure(cls, schema=None, name=None):
        cls.__table__.schema = schema
        cls.id.default.schema = schema
        cls.__table__.name = name or cls.__tablename__


class TaskExtended(Task):
    """For the extend result."""

    __tablename__ = 'celery_taskmeta'
    __table_args__ = {'sqlite_autoincrement': True, 'extend_existing': True}

    name = sa.Column(sa.String(155), nullable=True)
    args = sa.Column(sa.LargeBinary, nullable=True)
    kwargs = sa.Column(sa.LargeBinary, nullable=True)
    worker = sa.Column(sa.String(155), nullable=True)
    retries = sa.Column(sa.Integer, nullable=True)
    queue = sa.Column(sa.String(155), nullable=True)

    def to_dict(self):
        task_dict = super().to_dict()
        task_dict.update({
            'name': self.name,
            'args': self.args,
            'kwargs': self.kwargs,
            'worker': self.worker,
            'retries': self.retries,
            'queue': self.queue,
        })
        return task_dict


class TaskSet(ResultModelBase):
    """TaskSet result."""

    __tablename__ = 'celery_tasksetmeta'
    __table_args__ = {'sqlite_autoincrement': True}

    id = sa.Column(sa.Integer, sa.Sequence('taskset_id_sequence'),
                   autoincrement=True, primary_key=True)
    taskset_id = sa.Column(sa.String(155), unique=True)
    result = sa.Column(PickleType, nullable=True)
    date_done = sa.Column(sa.DateTime, default=datetime.utcnow,
                          nullable=True)

    def __init__(self, taskset_id, result):
        self.taskset_id = taskset_id
        self.result = result

    def to_dict(self):
        return {
            'taskset_id': self.taskset_id,
            'result': self.result,
            'date_done': self.date_done,
        }

    def __repr__(self):
        return f'<TaskSet: {self.taskset_id}>'

    @classmethod
    def configure(cls, schema=None, name=None):
        cls.__table__.schema = schema
        cls.id.default.schema = schema
        cls.__table__.name = name or cls.__tablename__



from collections import defaultdict
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.shortcuts import prompt
import click
import click._bashcomplete
import click.parser
import os
import shlex
import sys
import six
from .exceptions import InternalCommandException, ExitReplException  # noqa

# Handle click.exceptions.Exit introduced in Click 7.0
try:
    from click.exceptions import Exit as ClickExit
except ImportError:
    class ClickExit(RuntimeError):
        pass

PY2 = sys.version_info[0] == 2

if PY2:
    text_type = unicode  # noqa
else:
    text_type = str  # noqa


__version__ = "0.1.6"

_internal_commands = dict()


def _register_internal_command(names, target, description=None):
    if not hasattr(target, "__call__"):
        raise ValueError("Internal command must be a callable")

    if isinstance(names, six.string_types):
        names = [names]
    elif not isinstance(names, (list, tuple)):
        raise ValueError('"names" must be a string or a list / tuple')

    for name in names:
        _internal_commands[name] = (target, description)


def _get_registered_target(name, default=None):
    target_info = _internal_commands.get(name)
    if target_info:
        return target_info[0]
    return default


def _exit_internal():
    raise ExitReplException()


def _help_internal():
    formatter = click.HelpFormatter()
    formatter.write_heading("REPL help")
    formatter.indent()
    with formatter.section("External Commands"):
        formatter.write_text('prefix external commands with "!"')
    with formatter.section("Internal Commands"):
        formatter.write_text('prefix internal commands with ":"')
        info_table = defaultdict(list)
        for mnemonic, target_info in six.iteritems(_internal_commands):
            info_table[target_info[1]].append(mnemonic)
        formatter.write_dl(
            (
                ", ".join((":{0}".format(mnemonic) for mnemonic in sorted(mnemonics))),
                description,
            )
            for description, mnemonics in six.iteritems(info_table)
        )
    return formatter.getvalue()


_register_internal_command(["q", "quit", "exit"], _exit_internal, "exits the repl")
_register_internal_command(
    ["?", "h", "help"], _help_internal, "displays general help information"
)


class ClickCompleter(Completer):
    def __init__(self, cli):
        self.cli = cli

    def get_completions(self, document, complete_event=None):
        # Code analogous to click._bashcomplete.do_complete

        try:
            args = shlex.split(document.text_before_cursor)
        except ValueError:
            # Invalid command, perhaps caused by missing closing quotation.
            return

        cursor_within_command = (
            document.text_before_cursor.rstrip() == document.text_before_cursor
        )

        if args and cursor_within_command:
            # We've entered some text and no space, give completions for the
            # current word.
            incomplete = args.pop()
        else:
            # We've not entered anything, either at all or for the current
            # command, so give all relevant completions for this context.
            incomplete = ""

        ctx = click._bashcomplete.resolve_ctx(self.cli, "", args)
        if ctx is None:
            return

        choices = []
        for param in ctx.command.params:
            if isinstance(param, click.Option):
                for options in (param.opts, param.secondary_opts):
                    for o in options:
                        choices.append(
                            Completion(
                                text_type(o), -len(incomplete), display_meta=param.help
                            )
                        )
            elif isinstance(param, click.Argument):
                if isinstance(param.type, click.Choice):
                    for choice in param.type.choices:
                        choices.append(Completion(text_type(choice), -len(incomplete)))

        if isinstance(ctx.command, click.MultiCommand):
            for name in ctx.command.list_commands(ctx):
                command = ctx.command.get_command(ctx, name)
                choices.append(
                    Completion(
                        text_type(name),
                        -len(incomplete),
                        display_meta=getattr(command, "short_help"),
                    )
                )

        for item in choices:
            if item.text.startswith(incomplete):
                yield item


def bootstrap_prompt(prompt_kwargs, group):
    """
    Bootstrap prompt_toolkit kwargs or use user defined values.

    :param prompt_kwargs: The user specified prompt kwargs.
    """
    prompt_kwargs = prompt_kwargs or {}

    defaults = {
        "history": InMemoryHistory(),
        "completer": ClickCompleter(group),
        "message": u"> ",
    }

    for key in defaults:
        default_value = defaults[key]
        if key not in prompt_kwargs:
            prompt_kwargs[key] = default_value

    return prompt_kwargs


def repl(  # noqa: C901
    old_ctx,
    prompt_kwargs=None,
    allow_system_commands=True,
    allow_internal_commands=True,
):
    """
    Start an interactive shell. All subcommands are available in it.

    :param old_ctx: The current Click context.
    :param prompt_kwargs: Parameters passed to
        :py:func:`prompt_toolkit.shortcuts.prompt`.

    If stdin is not a TTY, no prompt will be printed, but only commands read
    from stdin.

    """
    # parent should be available, but we're not going to bother if not
    group_ctx = old_ctx.parent or old_ctx
    group = group_ctx.command
    isatty = sys.stdin.isatty()

    # Delete the REPL command from those available, as we don't want to allow
    # nesting REPLs (note: pass `None` to `pop` as we don't want to error if
    # REPL command already not present for some reason).
    repl_command_name = old_ctx.command.name
    if isinstance(group_ctx.command, click.CommandCollection):
        available_commands = {
            cmd_name: cmd_obj
            for source in group_ctx.command.sources
            for cmd_name, cmd_obj in source.commands.items()
        }
    else:
        available_commands = group_ctx.command.commands
    available_commands.pop(repl_command_name, None)

    prompt_kwargs = bootstrap_prompt(prompt_kwargs, group)

    if isatty:

        def get_command():
            return prompt(**prompt_kwargs)

    else:
        get_command = sys.stdin.readline

    while True:
        try:
            command = get_command()
        except KeyboardInterrupt:
            continue
        except EOFError:
            break

        if not command:
            if isatty:
                continue
            else:
                break

        if allow_system_commands and dispatch_repl_commands(command):
            continue

        if allow_internal_commands:
            try:
                result = handle_internal_commands(command)
                if isinstance(result, six.string_types):
                    click.echo(result)
                    continue
            except ExitReplException:
                break

        try:
            args = shlex.split(command)
        except ValueError as e:
            click.echo("{}: {}".format(type(e).__name__, e))
            continue

        try:
            with group.make_context(None, args, parent=group_ctx) as ctx:
                group.invoke(ctx)
                ctx.exit()
        except click.ClickException as e:
            e.show()
        except ClickExit:
            pass
        except SystemExit:
            pass
        except ExitReplException:
            break


def register_repl(group, name="repl"):
    """Register :func:`repl()` as sub-command *name* of *group*."""
    group.command(name=name)(click.pass_context(repl))


def exit():
    """Exit the repl"""
    _exit_internal()


def dispatch_repl_commands(command):
    """Execute system commands entered in the repl.

    System commands are all commands starting with "!".

    """
    if command.startswith("!"):
        os.system(command[1:])
        return True

    return False


def handle_internal_commands(command):
    """Run repl-internal commands.

    Repl-internal commands are all commands starting with ":".

    """
    if command.startswith(":"):
        target = _get_registered_target(command[1:], default=None)
        if target:
            return target()



import json
import re
import os
from threading import Thread
from celery import Celery
from functools import wraps
from flask_jwt_extended import get_jwt_identity
from itsdangerous import URLSafeTimedSerializer
from flask import url_for
from flask_mail import Mail, Message


class Validators:
    """Class with validation methods."""

    def __init__(self, variable=None):
        self.variable = variable

    def email(self):
        """Check if the format of the email is valid."""
        if re.match(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+[a-zA-Z0-9-.]+$)",
                    self.variable):
            return True
        return False

    def password(self):
        """Check if that the password is eight long character string with atleast one lowercase character, one uppercase character, one number, and one special character."""
        if re.match(r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)(?=.*[@$!%*?&])[A-Za-z\d@$!%*?&]{8,}$", self.variable):
            return True
        return False

    def phone(self):
        """Check if that the phone number satisfies kenyan format."""
        if re.match(r"^(?:254|\+254|0)?(7(?:(?:[129][0-9])|(?:0[0-8])|(4[0-1]))[0-9]{6})$", self.variable):
            return True
        return False

    def safaricom(self):
        """Check if that the phone number satisfies safaricom format."""
        if re.match(r"^(?:254|\+254|0)?(7(?:(?:[129][0–9])|(?:0[0–8])|(4[0–1]))[0–9]{6})$", self.variable):
            return True
        return False

    def airtel(self):
        """Check if that the phone number satisfies airtel format."""
        if re.match(r"^(?:254|\+254|0)?(7(?:(?:[3][0-9])|(?:5[0-6])|(8[0-9]))[0-9]{6})$", self.variable):
            return True
        return False

    def orange(self):
        """Check if that the phone number satisfies orange format."""
        if re.match(r"^(?:254|\+254|0)?(77[0-6][0-9]{6})$", self.variable):
            return True
        return False

    def equitel(self):
        """Check if that the phone number satisfies equitel format."""
        if re.match(r"^(?:254|\+254|0)?(76[34][0-9]{6})$", self.variable):
            return True
        return False

    def integer(self):
        """Check if the input is an integer."""
        if re.match(r"^[-+]?([1-9]\d*|0)$", self.variable):
            return True
        return False

    def name(self):
        """Check if the input is a word."""
        if re.match(r"^[A-Za-z]{2,25}||\s[A-Za-z]{2,25}$", self.variable):
            return True
        return False

    def error_in_json_keys(self, request):
        """Display errors if json keys don't match."""
        res_keys = self.variable
        errors = []
        for key in res_keys:
            if key not in request.json:
                errors.append(key)
        return errors

    def input_restriction(self, input_data):
        """Verify that the user input is what is expected."""
        form = input_data
        if self.variable not in form:
            return False
        return True


class Serializer:
    """This class serializes data."""

    @classmethod
    def serialize(cls, response, status_code, message=200):
        """Serializes data output."""
        if status_code in (400, 401, 403, 404, 405, 500):
            return json.dumps({
                "status": status_code,
                "message": message,
                "error": response
            }), status_code
        return json.dumps({
            "status": status_code,
            "message": message,
            "data": response
        }), status_code

    @classmethod
    def raise_error(cls, status, msg):
        """Display error message."""
        return json.dumps({
            "status": status,
            "message": msg
        }), status

    @classmethod
    def on_success(cls, status, msg):
        """Display successful message."""
        return json.dumps({
            "status": status,
            "message": msg
        }), status

    @classmethod
    def page_not_found(cls, e):
        """Capture not found error."""
        return json.dumps({
            "status": "404",
            "message": "resource not found"
        }), 404

    @classmethod
    def method_not_allowed(cls, e):
        """Capture method not allowed error."""
        return json.dumps({
            "status": "405",
            "message": "method not allowed"
        }), 405

    @classmethod
    def internal_server_error(cls, e):
        """Capture internal server error."""
        return json.dumps({
            "status": "500",
            "message": "internal server error"
        }), 500

    @classmethod
    def default_encode_token(cls, SECRET_KEY, variable, salt='default-key'):
        """Encode token using unique variable."""
        serializer = URLSafeTimedSerializer(SECRET_KEY)
        return serializer.dumps(variable, salt=salt)

    @classmethod
    def default_decode_token(cls, SECRET_KEY, token, salt='default-key', expiration=3600):
        """Decode token and get the variable."""
        serializer = URLSafeTimedSerializer(SECRET_KEY)
        try:
            variable = serializer.loads(
                token, salt=salt, max_age=expiration)
            return variable
        except Exception:
            return False

    @classmethod
    def generate_url(cls, endpoint, token):
        """Generate url to concatenate at the end of another url."""
        return url_for(endpoint, token=token, _external=True)


def admin_required(func, users):

    @ wraps(func)
    def wrapper_function(*args, **kwargs):

        try:
            cur_user = [
                user for user in users if user['email'] == get_jwt_identity()]
            user_role = cur_user[0]['role']
            if user_role != 'admin':
                return {
                    'message': 'This activity can only be completed by the admin'}, 403  # Forbidden
            return func(*args, **kwargs)
        except Exception as e:
            return {"message": e}

    return wrapper_function


def send_async_email(app, msg):
    """Send asychronous email."""
    mail = Mail()
    with app.app_context():
        try:
            mail.send(msg)
        except ConnectionRefusedError:
            return json.dumps({
                "status": "500",
                "msg": "Mail server not working"
            }), 500


def send_email(app, subject, sender, recipients, text_body, html_body):
    """Message body."""
    msg = Message(subject, sender=sender, recipients=recipients)
    msg.body = text_body
    msg.html = html_body
    Thread(target=send_async_email, args=(app, msg)).start()


def make_celery(app, broker_url, backend_url):
    celery = Celery(
        app.import_name,
        broker=broker_url,
        backend=backend_url
    )
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        def __call__(self, *args, **kwargs):
            with app.app_context():
                return self.run(*args, **kwargs)

    celery.Task = ContextTask
    return celery



"""App utilities: Compat settings, bug-report tool, pickling apps."""
import os
import platform as _platform
import re
from collections import namedtuple
from collections.abc import Mapping
from copy import deepcopy
from types import ModuleType

from kombu.utils.url import maybe_sanitize_url

from celery.exceptions import ImproperlyConfigured
from celery.platforms import pyimplementation
from celery.utils.collections import ConfigurationView
from celery.utils.imports import import_from_cwd, qualname, symbol_by_name
from celery.utils.text import pretty

from .defaults import (_OLD_DEFAULTS, _OLD_SETTING_KEYS, _TO_NEW_KEY,
                       _TO_OLD_KEY, DEFAULTS, SETTING_KEYS, find)

__all__ = (
    'Settings', 'appstr', 'bugreport',
    'filter_hidden_settings', 'find_app',
)

#: Format used to generate bug-report information.
BUGREPORT_INFO = """
software -> celery:{celery_v} kombu:{kombu_v} py:{py_v}
            billiard:{billiard_v} {driver_v}
platform -> system:{system} arch:{arch}
            kernel version:{kernel_version} imp:{py_i}
loader   -> {loader}
settings -> transport:{transport} results:{results}

{human_settings}
"""

HIDDEN_SETTINGS = re.compile(
    'API|TOKEN|KEY|SECRET|PASS|PROFANITIES_LIST|SIGNATURE|DATABASE',
    re.IGNORECASE,
)

E_MIX_OLD_INTO_NEW = """

Cannot mix new and old setting keys, please rename the
following settings to the new format:

{renames}

"""

E_MIX_NEW_INTO_OLD = """

Cannot mix new setting names with old setting names, please
rename the following settings to use the old format:

{renames}

Or change all of the settings to use the new format :)

"""

FMT_REPLACE_SETTING = '{replace:<36} -> {with_}'


def appstr(app):
    """String used in __repr__ etc, to id app instances."""
    return f'{app.main or "__main__"} at {id(app):#x}'


class Settings(ConfigurationView):
    """Celery settings object.

    .. seealso:

        :ref:`configuration` for a full list of configuration keys.

    """

    def __init__(self, *args, deprecated_settings=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.deprecated_settings = deprecated_settings

    @property
    def broker_read_url(self):
        return (
            os.environ.get('CELERY_BROKER_READ_URL') or
            self.get('broker_read_url') or
            self.broker_url
        )

    @property
    def broker_write_url(self):
        return (
            os.environ.get('CELERY_BROKER_WRITE_URL') or
            self.get('broker_write_url') or
            self.broker_url
        )

    @property
    def broker_url(self):
        return (
            os.environ.get('CELERY_BROKER_URL') or
            self.first('broker_url', 'broker_host')
        )

    @property
    def result_backend(self):
        return (
            os.environ.get('CELERY_RESULT_BACKEND') or
            self.first('result_backend', 'CELERY_RESULT_BACKEND')
        )

    @property
    def task_default_exchange(self):
        return self.first(
            'task_default_exchange',
            'task_default_queue',
        )

    @property
    def task_default_routing_key(self):
        return self.first(
            'task_default_routing_key',
            'task_default_queue',
        )

    @property
    def timezone(self):
        # this way we also support django's time zone.
        return self.first('timezone', 'time_zone')

    def without_defaults(self):
        """Return the current configuration, but without defaults."""
        # the last stash is the default settings, so just skip that
        return Settings({}, self.maps[:-1])

    def value_set_for(self, key):
        return key in self.without_defaults()

    def find_option(self, name, namespace=''):
        """Search for option by name.

        Example:
            >>> from proj.celery import app
            >>> app.conf.find_option('disable_rate_limits')
            ('worker', 'prefetch_multiplier',
             <Option: type->bool default->False>))

        Arguments:
            name (str): Name of option, cannot be partial.
            namespace (str): Preferred name-space (``None`` by default).
        Returns:
            Tuple: of ``(namespace, key, type)``.
        """
        return find(name, namespace)

    def find_value_for_key(self, name, namespace='celery'):
        """Shortcut to ``get_by_parts(*find_option(name)[:-1])``."""
        return self.get_by_parts(*self.find_option(name, namespace)[:-1])

    def get_by_parts(self, *parts):
        """Return the current value for setting specified as a path.

        Example:
            >>> from proj.celery import app
            >>> app.conf.get_by_parts('worker', 'disable_rate_limits')
            False
        """
        return self['_'.join(part for part in parts if part)]

    def finalize(self):
        # See PendingConfiguration in celery/app/base.py
        # first access will read actual configuration.
        try:
            self['__bogus__']
        except KeyError:
            pass
        return self

    def table(self, with_defaults=False, censored=True):
        filt = filter_hidden_settings if censored else lambda v: v
        dict_members = dir(dict)
        self.finalize()
        settings = self if with_defaults else self.without_defaults()
        return filt({
            k: v for k, v in settings.items()
            if not k.startswith('_') and k not in dict_members
        })

    def humanize(self, with_defaults=False, censored=True):
        """Return a human readable text showing configuration changes."""
        return '\n'.join(
            f'{key}: {pretty(value, width=50)}'
            for key, value in self.table(with_defaults, censored).items())

    def maybe_warn_deprecated_settings(self):
        # TODO: Remove this method in Celery 6.0
        if self.deprecated_settings:
            from celery.app.defaults import _TO_NEW_KEY
            from celery.utils import deprecated
            for setting in self.deprecated_settings:
                deprecated.warn(description=f'The {setting!r} setting',
                                removal='6.0.0',
                                alternative=f'Use the {_TO_NEW_KEY[setting]} instead')

            return True

        return False


def _new_key_to_old(key, convert=_TO_OLD_KEY.get):
    return convert(key, key)


def _old_key_to_new(key, convert=_TO_NEW_KEY.get):
    return convert(key, key)


_settings_info_t = namedtuple('settings_info_t', (
    'defaults', 'convert', 'key_t', 'mix_error',
))

_settings_info = _settings_info_t(
    DEFAULTS, _TO_NEW_KEY, _old_key_to_new, E_MIX_OLD_INTO_NEW,
)
_old_settings_info = _settings_info_t(
    _OLD_DEFAULTS, _TO_OLD_KEY, _new_key_to_old, E_MIX_NEW_INTO_OLD,
)


def detect_settings(conf, preconf=None, ignore_keys=None, prefix=None,
                    all_keys=None, old_keys=None):
    preconf = {} if not preconf else preconf
    ignore_keys = set() if not ignore_keys else ignore_keys
    all_keys = SETTING_KEYS if not all_keys else all_keys
    old_keys = _OLD_SETTING_KEYS if not old_keys else old_keys

    source = conf
    if conf is None:
        source, conf = preconf, {}
    have = set(source.keys()) - ignore_keys
    is_in_new = have.intersection(all_keys)
    is_in_old = have.intersection(old_keys)

    info = None
    if is_in_new:
        # have new setting names
        info, left = _settings_info, is_in_old
        if is_in_old and len(is_in_old) > len(is_in_new):
            # Majority of the settings are old.
            info, left = _old_settings_info, is_in_new
    if is_in_old:
        # have old setting names, or a majority of the names are old.
        if not info:
            info, left = _old_settings_info, is_in_new
        if is_in_new and len(is_in_new) > len(is_in_old):
            # Majority of the settings are new
            info, left = _settings_info, is_in_old
    else:
        # no settings, just use new format.
        info, left = _settings_info, is_in_old

    if prefix:
        # always use new format if prefix is used.
        info, left = _settings_info, set()

    # only raise error for keys that the user didn't provide two keys
    # for (e.g., both ``result_expires`` and ``CELERY_TASK_RESULT_EXPIRES``).
    really_left = {key for key in left if info.convert[key] not in have}
    if really_left:
        # user is mixing old/new, or new/old settings, give renaming
        # suggestions.
        raise ImproperlyConfigured(info.mix_error.format(renames='\n'.join(
            FMT_REPLACE_SETTING.format(replace=key, with_=info.convert[key])
            for key in sorted(really_left)
        )))

    preconf = {info.convert.get(k, k): v for k, v in preconf.items()}
    defaults = dict(deepcopy(info.defaults), **preconf)
    return Settings(
        preconf, [conf, defaults],
        (_old_key_to_new, _new_key_to_old),
        deprecated_settings=is_in_old,
        prefix=prefix,
    )


class AppPickler:
    """Old application pickler/unpickler (< 3.1)."""

    def __call__(self, cls, *args):
        kwargs = self.build_kwargs(*args)
        app = self.construct(cls, **kwargs)
        self.prepare(app, **kwargs)
        return app

    def prepare(self, app, **kwargs):
        app.conf.update(kwargs['changes'])

    def build_kwargs(self, *args):
        return self.build_standard_kwargs(*args)

    def build_standard_kwargs(self, main, changes, loader, backend, amqp,
                              events, log, control, accept_magic_kwargs,
                              config_source=None):
        return {'main': main, 'loader': loader, 'backend': backend,
                'amqp': amqp, 'changes': changes, 'events': events,
                'log': log, 'control': control, 'set_as_current': False,
                'config_source': config_source}

    def construct(self, cls, **kwargs):
        return cls(**kwargs)


def _unpickle_app(cls, pickler, *args):
    """Rebuild app for versions 2.5+."""
    return pickler()(cls, *args)


def _unpickle_app_v2(cls, kwargs):
    """Rebuild app for versions 3.1+."""
    kwargs['set_as_current'] = False
    return cls(**kwargs)


def filter_hidden_settings(conf):
    """Filter sensitive settings."""
    def maybe_censor(key, value, mask='*' * 8):
        if isinstance(value, Mapping):
            return filter_hidden_settings(value)
        if isinstance(key, str):
            if HIDDEN_SETTINGS.search(key):
                return mask
            elif 'broker_url' in key.lower():
                from kombu import Connection
                return Connection(value).as_uri(mask=mask)
            elif 'backend' in key.lower():
                return maybe_sanitize_url(value, mask=mask)

        return value

    return {k: maybe_censor(k, v) for k, v in conf.items()}


def bugreport(app):
    """Return a string containing information useful in bug-reports."""
    import billiard
    import kombu

    import celery

    try:
        conn = app.connection()
        driver_v = '{}:{}'.format(conn.transport.driver_name,
                                  conn.transport.driver_version())
        transport = conn.transport_cls
    except Exception:  # pylint: disable=broad-except
        transport = driver_v = ''

    return BUGREPORT_INFO.format(
        system=_platform.system(),
        arch=', '.join(x for x in _platform.architecture() if x),
        kernel_version=_platform.release(),
        py_i=pyimplementation(),
        celery_v=celery.VERSION_BANNER,
        kombu_v=kombu.__version__,
        billiard_v=billiard.__version__,
        py_v=_platform.python_version(),
        driver_v=driver_v,
        transport=transport,
        results=maybe_sanitize_url(app.conf.result_backend or 'disabled'),
        human_settings=app.conf.humanize(),
        loader=qualname(app.loader.__class__),
    )


def find_app(app, symbol_by_name=symbol_by_name, imp=import_from_cwd):
    """Find app by name."""
    from .base import Celery

    try:
        sym = symbol_by_name(app, imp=imp)
    except AttributeError:
        # last part was not an attribute, but a module
        sym = imp(app)
    if isinstance(sym, ModuleType) and ':' not in app:
        try:
            found = sym.app
            if isinstance(found, ModuleType):
                raise AttributeError()
        except AttributeError:
            try:
                found = sym.celery
                if isinstance(found, ModuleType):
                    raise AttributeError("attribute 'celery' is the celery module not the instance of celery")
            except AttributeError:
                if getattr(sym, '__path__', None):
                    try:
                        return find_app(
                            f'{app}.celery',
                            symbol_by_name=symbol_by_name, imp=imp,
                        )
                    except ImportError:
                        pass
                for suspect in vars(sym).values():
                    if isinstance(suspect, Celery):
                        return suspect
                raise
            else:
                return found
        else:
            return found
    return sym



"""Abstract types."""
import abc
import asyncio
import socket
from array import array
from datetime import datetime
from typing import (
    Any, Awaitable, Callable, IO, List, Mapping,
    MutableMapping, NamedTuple, Optional, Sequence, SupportsInt,
    TypeVar, Tuple, Union,
)
from .protocol import queue_declare_ok_t
from .spec import method_sig_t
from vine import Thenable

Fd = TypeVar('Fd', int, IO)
Int = TypeVar('Int', SupportsInt, str)


class Frame(NamedTuple):
    type: int
    channel: int
    data: bytes


ConnectionBlockedCallbackT = Callable[[str], Optional[Awaitable]]
ConnectionUnblockedCallbackT = Callable[[], Optional[Awaitable]]
ConnectionFrameHandlerT = Callable[[Frame], Awaitable]
ConnectionFrameWriterT = Callable[
    [int,
     int,
     Optional[method_sig_t],
     Optional[bytes],
     Optional['MessageT'],
     Optional[float]],
    Awaitable,
]

WaitMethodT = Union[method_sig_t, Sequence[method_sig_t]]


class TransportT(metaclass=abc.ABCMeta):
    """Transport type."""

    rstream: asyncio.StreamReader
    wstream: asyncio.StreamWriter

    connected: bool = False
    host: str
    port: int
    ssl: Any
    connect_timeout: float
    read_timeout: float
    write_timeout: float
    socket_settings: Mapping
    sock: socket.socket

    @abc.abstractmethod
    async def connect(self) -> None:
        ...

    @abc.abstractmethod
    def close(self) -> None:
        ...

    @abc.abstractmethod
    async def read_frame(self, timeout: float = None) -> Frame:
        ...

    @abc.abstractmethod
    async def write(self, s: bytes, timeout: float = None) -> None:
        ...


class ContentT(metaclass=abc.ABCMeta):
    """Generic content type."""

    CLASS_ID: int
    PROPERTIES: Sequence[Tuple[str, str]]

    properties: MutableMapping
    body_received: int = 0
    body_size: int = 0
    ready: bool = False

    frame_method: method_sig_t
    frame_args: str

    @abc.abstractmethod
    def _load_properties(
            self,
            class_id: int,
            buf: bytes,
            offset: int = 0,
            classes: Mapping = None,
            unpack_from: Callable = None) -> int:
        ...

    @abc.abstractmethod
    def _serialize_properties(self) -> bytes:
        ...

    @abc.abstractmethod
    def inbound_header(self, buf: bytes, offset: int = 0) -> int:
        ...

    @abc.abstractmethod
    def inbound_body(self, buf: bytes):
        ...


class MessageT(ContentT, metaclass=abc.ABCMeta):
    """Basic message type."""

    body: bytes
    children: Any
    channel: 'ChannelT'
    delivery_info: Mapping[str, Any]

    content_type: str
    content_encoding: str
    application_headers: MutableMapping
    delivery_mode: int
    priority: int
    correlation_id: str
    reply_to: str
    expiration: str
    message_id: str
    timestamp: datetime
    type: str
    user_id: str
    app_id: str
    cluster_id: str

    @abc.abstractmethod
    def __init__(self,
                 body: bytes=b'',
                 *,
                 children: Any = None,
                 channel: 'ChannelT' = None) -> None:
        ...

    @property
    @abc.abstractmethod
    def headers(self) -> MutableMapping:
        ...

    @property
    @abc.abstractmethod
    def delivery_tag(self) -> str:
        ...


class AbstractChannelT(metaclass=abc.ABCMeta):
    """Abstract channel type."""

    connection: 'ConnectionT'
    channel_id: int
    auto_decode: bool = False
    is_open: bool = False

    @abc.abstractmethod
    def __enter__(self) -> 'AbstractChannelT':
        ...

    @abc.abstractmethod
    def __exit__(self, *exc_info) -> None:
        ...

    @abc.abstractmethod
    async def __aenter__(self) -> 'AbstractChannelT':
        ...

    @abc.abstractmethod
    async def __aexit__(self, *exc_info) -> None:
        ...

    @abc.abstractmethod
    def _setup_listeners(self):
        ...

    @abc.abstractmethod
    async def send_method(
            self, sig: method_sig_t,
            format: str = None,
            args: Sequence = None,
            *,
            content: MessageT = None,
            wait: WaitMethodT = None,
            callback: Callable = None,
            returns_tuple: bool = False) -> Thenable:
        ...

    @abc.abstractmethod
    async def close(
            self,
            *,
            reply_code: int = 0,
            reply_text: str = '',
            method_sig: method_sig_t = method_sig_t(0, 0),
            argsig: str = 'BsBB') -> None:
        ...

    @abc.abstractmethod
    def collect(self) -> None:
        ...

    @abc.abstractmethod
    async def wait(
            self,
            method: WaitMethodT,
            *,
            callback: Callable = None,
            timeout: float = None,
            returns_tuple: bool = False) -> Any:
        ...

    async def dispatch_method(
            self,
            method_sig: method_sig_t,
            payload: bytes,
            content: MessageT) -> None:
        ...


class ConnectionT(AbstractChannelT):
    """Connection channel type."""

    Channel: type
    Transport: type

    host: str
    userid: str
    password: str
    login_method: str
    login_response: Any
    virtual_host: str
    locale: str
    client_properties: MutableMapping
    ssl: Any
    channel_max: int
    frame_max: int
    on_open: Thenable
    on_tune_ok: Thenable
    confirm_publish: bool
    connect_timeout: float
    read_timeout: float
    write_timeout: float
    socket_settings: Mapping

    negotiate_capabilities: Mapping[str, bool]
    library_properties: Mapping[str, Any]
    heartbeat: float
    client_heartbeat: float
    server_heartbeat: float
    last_heartbeat_sent: float
    last_heartbeat_received: float
    bytes_sent: int = 0
    bytes_recv: int = 0
    prev_sent: int
    prev_recv: int

    connection_errors: Tuple[type, ...]
    channel_errors: Tuple[type, ...]
    recoverable_connection_errors: Tuple[type, ...]
    recoverable_channel_errors: Tuple[type, ...]

    transport: TransportT
    channels: MutableMapping[int, AbstractChannelT]
    loop: asyncio.AbstractEventLoop

    mechanisms: List[str]
    locales: List[str]

    _avail_channel_ids: array

    def __init__(
            self,
            host: str = 'localhost:5672',
            userid: str = 'guest',
            password: str = 'guest',
            *,
            login_method: str = 'AMQPLAIN',
            login_response: Any = None,
            virtual_host: str = '/',
            locale: str = 'en_US',
            client_properties: Mapping = None,
            ssl: Any = False,
            connect_timeout: float = None,
            channel_max: int = None,
            frame_max: int = None,
            heartbeat: float = 0.0,
            on_open: Thenable = None,
            on_blocked: ConnectionBlockedCallbackT = None,
            on_unblocked: ConnectionUnblockedCallbackT = None,
            confirm_publish: bool = False,
            on_tune_ok: Callable = None,
            read_timeout: float = None,
            write_timeout: float = None,
            socket_settings: Mapping = None,
            frame_handler: ConnectionFrameHandlerT = None,
            frame_writer: ConnectionFrameWriterT = None,
            loop: asyncio.AbstractEventLoop = None,
            transport: TransportT = None,
            **kwargs) -> None:
        self.frame_writer = frame_writer
        self.frame_handler = frame_handler

    @property
    @abc.abstractmethod
    def server_capabilities(self) -> Mapping:
        ...

    @property
    @abc.abstractmethod
    def sock(self) -> socket.socket:
        pass

    @abc.abstractmethod
    async def connect(self, callback: Callable[[], None] = None) -> None:
        ...

    @property
    @abc.abstractmethod
    def connected(self) -> bool:
        ...

    @abc.abstractmethod
    def channel(self, channel_id: int,
                callback: Callable = None) -> 'AbstractChannelT':
        ...

    @abc.abstractmethod
    def is_alive(self) -> bool:
        ...

    @abc.abstractmethod
    async def drain_events(self, timeout: float = None) -> None:
        ...

    @abc.abstractmethod
    async def on_inbound_method(
            self,
            channel_id: int,
            method_sig: method_sig_t,
            payload: bytes,
            content: MessageT) -> None:
        ...

    @abc.abstractmethod
    async def send_heartbeat(self) -> None:
        ...

    @abc.abstractmethod
    async def heartbeat_tick(self, rate: int = 2) -> None:
        ...

    def _get_free_channel_id(self) -> int:
        ...

    @abc.abstractmethod
    def _claim_channel_id(self, channel_id: int) -> None:
        ...


class ChannelT(AbstractChannelT, metaclass=abc.ABCMeta):
    """Channel type."""

    @abc.abstractmethod
    async def flow(self, active: bool) -> None:
        ...

    @abc.abstractmethod
    async def open(self) -> None:
        ...

    @abc.abstractmethod
    async def exchange_declare(
            self, exchange: str, type: str,
            *,
            passive: bool = False,
            durable: bool = False,
            auto_delete: bool = True,
            nowait: bool = False,
            arguments: Mapping[str, Any] = None,
            argsig: str = 'BssbbbbbF') -> None:
        ...

    @abc.abstractmethod
    async def exchange_delete(
            self, exchange: str,
            *,
            if_unused: bool = False,
            nowait: bool = False,
            argsig: str = 'Bsbb') -> None:
        ...

    @abc.abstractmethod
    async def exchange_bind(
            self, destination: str,
            source: str = '',
            routing_key: str = '',
            *,
            nowait: bool = False,
            arguments: Mapping[str, Any] = None,
            argsig: str = 'BsssbF') -> None:
        ...

    @abc.abstractmethod
    async def exchange_unbind(
            self, destination: str,
            source: str = '',
            routing_key: str = '',
            *,
            nowait: bool = False,
            arguments: Mapping[str, Any] = None,
            argsig: str = 'BsssbF') -> None:
        ...

    @abc.abstractmethod
    async def queue_bind(
            self, queue: str,
            exchange: str = '',
            routing_key: str = '',
            *,
            nowait: bool = False,
            arguments: Mapping[str, Any] = None,
            argsig: str = 'BsssbF') -> None:
        ...

    @abc.abstractmethod
    async def queue_unbind(
            self, queue: str, exchange: str,
            routing_key: str = '',
            *,
            nowait: bool = False,
            arguments: Mapping[str, Any] = None,
            argsig: str = 'BsssF') -> None:
        ...

    @abc.abstractmethod
    async def queue_declare(
            self,
            queue: str = '',
            *,
            passive: bool = False,
            durable: bool = False,
            exclusive: bool = False,
            auto_delete: bool = True,
            nowait: bool = False,
            arguments: Mapping[str, Any] = None,
            argsig: str = 'BsbbbbbF') -> queue_declare_ok_t:
        ...

    @abc.abstractmethod
    async def queue_delete(
            self,
            queue: str = '',
            *,
            if_unused: bool = False,
            if_empty: bool = False,
            nowait: bool = False,
            argsig: str = 'Bsbbb') -> None:
        ...

    @abc.abstractmethod
    async def queue_purge(
            self,
            queue: str = '',
            *,
            nowait: bool = False,
            argsig: str = 'Bsb') -> Optional[int]:
        ...

    @abc.abstractmethod
    async def basic_ack(
            self, delivery_tag: str,
            *,
            multiple: bool = False,
            argsig: str = 'Lb') -> None:
        ...

    @abc.abstractmethod
    async def basic_cancel(
            self, consumer_tag: str,
            *,
            nowait: bool = False,
            argsig: str = 'sb') -> None:
        ...

    @abc.abstractmethod
    async def basic_consume(
            self,
            queue: str = '',
            consumer_tag: str = '',
            *,
            no_local: bool = False,
            no_ack: bool = False,
            exclusive: bool = False,
            nowait: bool = False,
            callback: Callable = None,
            arguments: Mapping[str, Any] = None,
            on_cancel: Callable = None,
            argsig: str = 'BssbbbbF') -> None:
        ...

    @abc.abstractmethod
    async def basic_get(
            self,
            queue: str = '',
            *,
            no_ack: bool = False,
            argsig: str = 'Bsb') -> Optional[MessageT]:
        ...

    @abc.abstractmethod
    async def basic_publish(
            self, msg: MessageT,
            exchange: str = '',
            routing_key: str = '',
            *,
            mandatory: bool = False,
            immediate: bool = False,
            timeout: float = None,
            argsig: str = 'Bssbb') -> None:
        ...

    @abc.abstractmethod
    async def basic_qos(
            self,
            prefetch_size: int,
            prefetch_count: int,
            a_global: bool,
            argsig: str = 'lBb') -> None:
        ...

    @abc.abstractmethod
    async def basic_recover(self, *, requeue: bool = False) -> None:
        ...

    @abc.abstractmethod
    async def basic_recover_async(self, *, requeue: bool = False) -> None:
        ...

    @abc.abstractmethod
    async def basic_reject(self, delivery_tag: str,
                           *,
                           requeue: bool = False,
                           argsig: str = 'Lb') -> None:
        ...

    @abc.abstractmethod
    async def tx_commit(self) -> None:
        ...

    @abc.abstractmethod
    async def tx_rollback(self) -> None:
        ...

    @abc.abstractmethod
    async def tx_select(self) -> None:
        ...

    @abc.abstractmethod
    async def confirm_select(self, *, nowait: bool = False) -> None:
        ...


"""Entry-point for the :program:`celery` umbrella command."""

import sys

from . import maybe_patch_concurrency

__all__ = ('main',)


def main():
    """Entrypoint to the ``celery`` umbrella command."""
    if 'multi' not in sys.argv:
        maybe_patch_concurrency()
    from celery.bin.celery import main as _main
    sys.exit(_main())


if __name__ == '__main__':  # pragma: no cover
    main()



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
 
classifiers = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Education',
  'Operating System :: OS Independent',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3',
  'Programming Language :: Python :: 3.6',
  'Programming Language :: Python :: 3.7'
]

# python3 setup.py sdist bdist_wheel
# twine upload --skip-existing dist/*
 
setup(
  name='arrotechtools',
  version='0.0.8',
  description='Library with most of the useful methods in the world of programming',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url='',  
  author='Harun Gachanja',
  author_email='arrotechdesign@gmail.com',
  license='MIT', 
  classifiers=classifiers,
  keywords='roles', 
  packages=find_packages(),
  install_requires=['Flask-JWT-Extended==4.1.0', 'celery==5.0.5', 'Flask==1.1.2', 'Flask-Mail==0.9.1'] 
)